//----------------------------------------------------------------------
/// 
/// Boosted Higgs tagging
///
/// Template Overlap example program, illustration of carrying out boosted
/// Higgs subjet identification analysis
///
/// It illustrates two kinds of functionality: 
///
///  - using a boosted higgs tagger, that combines both 2-body and 3-body templates
///  - following information from jet shapes, such as planar flow
///
/// This kind of functionality was used in arXiv:1212.2977 and arXiv:1112.1957
///  for boosted Higgs searches,
/// and related functionality was used in arXiv:1211.2202 (ATLAS) in searching for boosted tops
///
/// run it with    : ./example5 WH8TeV.dat
///
/// Source code: example5.cc
//----------------------------------------------------------------------

#include "matching.hh"
#include "TemplateTagger.hh"

#include "fastjet/FunctionOfPseudoJet.hh"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <time.h> // For timing
#include <iomanip> 
#include <stdio.h> // For pipe

#include "Settings.hh"



using namespace std;

using namespace TemplateOverlap;

unsigned int WhichPtBin (double PtJet, double PtMin, double PtBinWidth, int nptbins)
{
if (PtJet < PtMin) return 0;
int i =  int(floor  ((PtJet - PtMin)/PtBinWidth));
if (i > nptbins) i = nptbins;
if (i<0) return 0;
return i;
}



///  Helper function for printing jets
void PrintJets(const vector <fastjet::PseudoJet>& jets) {
   printf("%5s %15s %15s %15s %15s %15s\n","jet #", "rapidity", "phi", "pt","m","e"); // label the columns
   for (unsigned int i = 0; i < jets.size(); i++) {
      printf("%5u %15.8f %15.8f %15.8f %15.8f %15.8f\n",i, jets[i].rap(),jets[i].phi(),jets[i].perp(),jets[i].m(),jets[i].e());
   }
}


//----------------------------------------------------------------------
// forward declaration for printing out info about a jet
//----------------------------------------------------------------------
ostream & operator<<(ostream &, const fastjet::PseudoJet &);


//----------------------------------------------------------------------
// class for reading in a jet
//----------------------------------------------------------------------

bool ReadEvent (ifstream& is,  
		vector<fastjet::PseudoJet> & input_particles, 
		vector<fastjet::PseudoJet> & hard_partons)
{

  string line;
  double pup1, pup2, pup3, pup4;
  int nhep, i, isthep,idhep,evtnumber;
  
  input_particles.resize(0);
  hard_partons.resize(0);
  
  if (!getline(is,line)) return false;	
  istringstream getpro(line);
  getpro  >> evtnumber >> nhep;

  for (int j=0; j<nhep;++j){
    if (!getline(is,line)) return false;	
    istringstream getall(line);
    getall  >> i >> isthep >>idhep>> pup1 >> pup2 >> pup3 >> pup4;
	
    if ( isthep == 21 )
    {
      fastjet::PseudoJet pJet = fastjet::PseudoJet(pup1,pup2,pup3,pup4);
      pJet.set_user_index(idhep);  
      hard_partons.push_back(pJet);
    } 

    if (fabs(idhep) == 12 || fabs(idhep) == 14 || fabs(idhep) == 16) continue;
    
     if ( isthep == 1 )
    {
      fastjet::PseudoJet pJet = fastjet::PseudoJet(pup1,pup2,pup3,pup4);
      pJet.set_user_index(idhep);  
      input_particles.push_back(pJet);
    } 
  } 

return evtnumber;

}

void splitPartons(const std::vector<fastjet::PseudoJet> & partons, std::vector<fastjet::PseudoJet> & bquarks, std::vector<fastjet::PseudoJet> & leptons)
{
  bquarks.resize(0);
  leptons.resize(0);
  for (unsigned int i=0;i<partons.size();++i)
  {
    int id = abs(partons[i].user_index());
    if (id==5) bquarks.push_back(partons[i]);
    if (id>=11 && id<=14) leptons.push_back(partons[i]);
  }
  
}

//---------------------------------------------------------
/////////////////Add pileup
//---------------------------------------------------------

/// This adds a fix number of pp interactions, useful when making plots
void addPileup(std::vector<fastjet::PseudoJet> & particles, std::ifstream & pileupFile, const int nPileup)
{
  std::vector <fastjet::PseudoJet> leptons, pileups;
  for (int iPileup = 0; iPileup < nPileup; ++iPileup) 
  {
    if(!ReadEvent(pileupFile,pileups, leptons)) 
    {
      pileupFile.clear() ;
      pileupFile.seekg(0, ios::beg) ;
      ReadEvent(pileupFile,pileups, leptons);
      
    }
    for (unsigned int i = 0; i < pileups.size(); ++i)
      particles.push_back(pileups[i]);
  }
}

int main(int argc, char* argv[]) {
  
  if (argc !=2) {
    cout << "Usage:  "<< argv[0] << " <input.dat>" << endl;
    return 1;
  }
  ///-----------------------------------------------------------------
  /// Read parameters from param.dat file
  /// Many of these are not useful in this example
  ///----------------------------------------------------------------
  Settings settings;
  settings.readFile("param.dat") ;
 
  double coneRadius =  settings.parm("coneRadius"), 
      coneRadius2 =  settings.parm("coneRadius2"), 
      subConeRadius2 =  settings.parm("subConeRadius2"), 
      subConeRadius3 = settings.parm("subConeRadius3"),
      subConeComparison2 = settings.parm("subConeComparison2"),
      sigma = settings.parm("sigma");
  bool VariableCone = settings.mode("variablecone");
  double pTbinWidth = settings.parm ("ptbinwidth");
	
  double nPileupAvg = settings.parm("nPileupAvg"), minPt_clus = settings.parm("minPtclus");
  double minPtW = settings.parm("minPtW"), minMET = settings.parm("minMET");	
  double minPt = settings.parm("minPt"), 
      maxPt =settings.parm("maxPt"), 
      minMass = settings.parm("minmass"), 
      maxMass = settings.parm("maxmass"); 
  int njetoutside = settings.mode("njetoutside");
    

  double ptjetmin = settings.parm("ptjetmin");
  int nptbins = settings.mode("nptbins");

  string fileName2[] = {settings.word("file2b1"),
      settings.word("file2b2"),
      settings.word("file2b3"),
      settings.word("file2b4"),
      settings.word("file2b5"),
      settings.word("file2b6"),            
      settings.word("file2b7")};            		    
  string fileName3[] = {settings.word("file3b1"),
      settings.word("file3b2"),
      settings.word("file3b3"),
      settings.word("file3b4"),
      settings.word("file3b5"),
      settings.word("file3b6"),
      settings.word("file3b7")};

  ///--------------------------------------------------------------------------------

  
   /// Set Run Parameters 
  double ovcut = 0.; /// For TemplateTagger
  int nPileup = 0; /// For pileup overlay
 
 /// Create best matched template containers
  std::vector<fastjet::PseudoJet> maxTempl;

 /// Create event containers
  std::vector<fastjet::PseudoJet> input_particles, partons;
  std::vector<fastjet::PseudoJet> myHiggs, myW;

 /// Where to read the events from
  ifstream file(argv[1]);
  ifstream pileupFile("pileup.dat");
 
 /// create 7 taggers (2-body) and 7 filters (3-body)
 ///----------------------------------------------------------
  std::vector<fastjet::Transformer *> taggers;
  std::vector<fastjet::Transformer *> filters;
  
  TemplateOverlap::TemplateDefinition templ2bDef(subConeRadius2, sigma, FIXED, CONE, ovcut);
  TemplateOverlap::TemplateDefinition templ3bDef(subConeRadius3, sigma, VARIABLE, CONE, ovcut);
  
  for( int i=0; i<nptbins;++i)
  {
    taggers.push_back(new fastjet::TemplateTagger(fileName2[i], templ2bDef, fastjet::SelectorMassRange(110,250)));
    filters.push_back(new fastjet::TemplateTagger(fileName3[i], templ3bDef, fastjet::SelectorMassRange(110,250)));
  }
 
  /// read in data in format i isthep idhep px py pz E 

  while(ReadEvent(file, input_particles, partons)){
   clock_t start = clock(); // start clock

   /// Jet definitions 
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::antikt_algorithm, coneRadius,
                                      fastjet::E_scheme, fastjet::Best);
  /// Overlay pileup
  if (nPileup > 0) addPileup(input_particles, pileupFile, nPileup);
  
   /// Find antikt jets
  vector <fastjet::PseudoJet> fatjets;
  fastjet::ClusterSequence clustSeq(input_particles, *jetDef);    
  fatjets = sorted_by_pt(clustSeq.inclusive_jets(25.));
  
  /// Print jet information 
  cout << "Run: " << jetDef->description() << endl << endl;
  cout << "Hardest jet: " << fatjets[0] << endl << endl;
  
  /// Choose a tagger (and a filter) based on jet pT
  int iTemp = WhichPtBin(fatjets[0].perp(), minPtW, pTbinWidth, nptbins-1); 
  const fastjet::Transformer & f2 = *taggers[iTemp];
  const fastjet::Transformer & f3 = *filters[iTemp]; 
  std::cout << "Using Template Library "<< iTemp << std::endl;
  
  /// Now tag the leading jet using template tagger
  /// For the Higgs, we combine 2-body and 3-body templates
  fastjet::PseudoJet tagged = f2(fatjets[0]);
  fastjet::PseudoJet filtered = f3(fatjets[0]);
  
  
  splitPartons(partons, myHiggs, myW);


  /// Show me what you got
  if (tagged == 0)
  {
    std::cout << "No 2-body substructure found" << std::endl;
  }  
  else
  {
    /// We extract substructure information from tagged jet  
    const fastjet::TemplateTagger::StructureType & templ2b = tagged.structure_of<fastjet::TemplateTagger>(); 
  
    fastjet::PseudoJet parent1 = tagged.pieces()[0];
    fastjet::PseudoJet parent2 = tagged.pieces()[1];
    std::cout << "Found suitable pair of subjets: " << std::endl;
    std::cout << " " << parent1 << std::endl;
    std::cout << " " << parent2 << std::endl;
    std::cout << "Total = " << std::endl;
    std::cout << " " << tagged << std::endl;
    std::cout << "(Ov2 = " << templ2b.ov() <<")" << std::endl << std::endl;
    std::cout << " The best-matched (2-body) templates are: "<< std::endl;     
    PrintJets(templ2b.maxTempl());
  } 
  
  /// Show me what you got
  if (filtered == 0)
  {
    std::cout << "No 3-body substructure found" << std::endl;
  }  
  else
  {
  
   /// We extract substructure information from filtered jet  
            
  const fastjet::TemplateTagger::StructureType & templ3b = filtered.structure_of<fastjet::TemplateTagger>(); 
  
  fastjet::PseudoJet parent1 = filtered.pieces()[0];
  fastjet::PseudoJet parent2 = filtered.pieces()[1];
  fastjet::PseudoJet parent3 = filtered.pieces()[2];

  std::cout << "Found suitable filtered subjets: " << std::endl;
  std::cout << " " << parent1 << std::endl;
  std::cout << " " << parent2 << std::endl;
  std::cout << " " << parent3 << std::endl;

  std::cout << "Total = " << std::endl;
  std::cout << " " << filtered << std::endl;
  std::cout << "(Ov3 = " << templ3b.ov() <<")" << std::endl << std::endl;
  std::cout << " The best-matched (3-body) template is: "<< std::endl;     
  PrintJets(templ3b.maxTempl());
  
  // Print additional information
  // Below we show how to compute planar flow 
  // directly from the jet constituents
  // or, alternatively, from the template partons
  
  // Extract the peak 3-body template
  std::vector<fastjet::PseudoJet> my3bTemplate = templ3b.maxTempl();
  
  // Calculate the template radius from template parton pT
  double r1 = eShape(my3bTemplate[0].perp()) - eShape(100.) + subConeRadius3;
  double r2 = eShape(my3bTemplate[1].perp()) - eShape(100.) + subConeRadius3;
  double r3 = eShape(my3bTemplate[2].perp()) - eShape(100.) + subConeRadius3;
 
  // Define two planar flow functions
  // This definition of planar flow smears the energy according to the template subcones
  fastjet::planarflow pf(r1,r2,r3);
  // This computes the actual planar flow
  fastjet::planarflow pf2(0,0,0);
  
  std::cout << "3-body template radii " << r1 << " "<< r2 << " "<< r3 << std::endl;
  
  // Create a simple jet with 3 particles out of template momenta
  fastjet::PseudoJet axis3b = fastjet::join(my3bTemplate);
  // Compute the actual jet planar flow and the template planar flow
  cout << "Jet Pf = " << pf2(fatjets[0]) << std::endl;
  cout << " Template Pf = " << pf(axis3b) << std::endl; 
  }
 
 // Show me how long it took
  std::cout << "Time Elapsed: " << ((double)clock() - start) / CLOCKS_PER_SEC << std::endl;

  std::string mystr;
  std::cout << "Please enter (y) to continue or (q) to abort the program:  ";
  std::getline (cin, mystr);
  if (mystr == "q") return 0;
  
  }
  
  // a bit of memory cleaning
  for (unsigned int i=0; i < taggers.size(); i++) delete taggers[i];
  for (unsigned int i=0; i < filters.size(); i++) delete filters[i];
  return 0;
}


//----------------------------------------------------------------------
// does the actual work for printing out a jet
//----------------------------------------------------------------------
ostream & operator<<(ostream & ostr, const fastjet::PseudoJet & jet) {
  ostr << "pt, y, phi =" 
       << " " << setw(10) << jet.perp() 
       << " " << setw(6) <<  jet.rap()  
       << " " << setw(6) <<  jet.phi()  
       << ", mass = " << setw(10) << jet.m()
       << ", btag = " << jet.user_index();
  return ostr;
	
	
}
